// Copyright (c) 2023 VMware Carbon Black, all rights reserved
////////////////////////////////////////////////////////////////////////////////
// These tests validate the pact between the Hello service (provider) and the
// Last Contact Time Updater service (consumer), specifically involving
// the insertion of data on the kinesis stream that the LCT will digest.
////////////////////////////////////////////////////////////////////////////////

package test

import (
	"fmt"
	"os"
	"testing"

	"hello-api/internal/lct"

	"carbonblack.com/protocols/psc/devicelastcontacttime"
	"github.com/pact-foundation/pact-go/v2/message"
	"github.com/pact-foundation/pact-go/v2/models"
	"github.com/pact-foundation/pact-go/v2/provider"
	"github.com/pact-foundation/pact-go/v2/version"
	"github.com/stretchr/testify/assert"
)

var localDevIndicator = "LOCALDEV"

func TestLctPublisherPact(t *testing.T) {
	version.CheckVersion()
	verifier := provider.NewVerifier()

	providerVersion := getVersion()
	providerBranch := getBranch()
	isLocal := providerVersion == localDevIndicator && providerBranch == localDevIndicator

	pactFiles, brokerURL := getPactFileLocation()

	handlerInvocationCount := map[string]int{}
	lctMessageDescription := "lct message"
	malformedLctMessageDescription := "malformed lct message"

	// Verify the Provider with broker
	err := verifier.VerifyProvider(t, provider.VerifyRequest{
		Provider:                   "ems-device-hello",
		ProviderBranch:             providerBranch,
		ProviderVersion:            providerVersion,
		PactFiles:                  pactFiles,
		BrokerURL:                  brokerURL,
		PublishVerificationResults: !isLocal,
		ConsumerVersionSelectors: []provider.Selector{
			&provider.ConsumerVersionSelector{
				MainBranch: true,
			},
		},
		MessageHandlers: message.Handlers{
			lctMessageDescription: func(s []models.ProviderState) (message.Body, message.Metadata, error) {
				handlerInvocationCount[lctMessageDescription]++
				lctRecord := &devicelastcontacttime.DeviceLastContactTime{
					OrgKey:               "7ZP4H5XR",
					DeviceId:             121,
					LastContactTimestamp: 121,
					DeregisterAfterTime:  121,
				}
				protobufBytes, _ := lct.FormatLCTMessage(lctRecord)
				return protobufBytes, message.Metadata{"contentType": "*/*"}, nil
			},

			malformedLctMessageDescription: func(s []models.ProviderState) (message.Body, message.Metadata, error) {
				handlerInvocationCount[malformedLctMessageDescription]++
				return []byte("DEFINETLYNOTGOODLOL"), message.Metadata{"contentType": "*/*"}, nil
			},
		},
	})

	assert.Equal(t, 2, len(handlerInvocationCount))
	assert.Equal(t, 1, handlerInvocationCount[lctMessageDescription])
	assert.Equal(t, 1, handlerInvocationCount[malformedLctMessageDescription])
	assert.NoError(t, err)
}

func getVersion() string {
	providerVersion := localDevIndicator
	if sha, ok := os.LookupEnv("CI_COMMIT_SHORT_SHA"); ok {
		providerVersion = fmt.Sprintf("1.0.0-SHA%s", sha)
	}

	return providerVersion
}

func getBranch() string {
	providerBranch := localDevIndicator
	if ciBranch, ok := os.LookupEnv("CI_COMMIT_SHORT_SHA"); ok {
		providerBranch = ciBranch
	}

	return providerBranch
}

func getPactFileLocation() (pactFiles []string, brokerURL string) {
	pactFiles = []string{}
	brokerURL = "https://pact-broker.mosaic-dev.sre-tools.carbonblack.io"

	if _, ok := os.LookupEnv("USE_LOCAL_PACT_FILE"); ok {
		pactFiles = []string{"pacts/ems-last-contact-updater-ems-device-hello.json"}
		brokerURL = ""
	}

	return pactFiles, brokerURL
}
