#  coding: utf-8
#  Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved
################################################################################

import json
import logging
from typing import Dict, Optional

import requests

_logger = logging.getLogger(__name__)


class BaseConnect(object):
    """
    Core class to handle basic rest calls and responses.
    """

    HTTPError = requests.HTTPError

    # todo: handle https

    def __init__(self, ip: str, port: int):

        if not isinstance(ip, str):
            raise TypeError("`ip` must be a str!")
        if not isinstance(port, int):
            raise TypeError("`port` must be an int!")

        self.protocol = "http"
        self.ip = ip
        self.port = port

        _logger.debug(f"Created connection to {self.protocol}://{self.ip}:{self.port}")
        self.session = requests.Session()
        self.session.headers.update({'User-Agent': 'Hinting Functional Tests', 'Connection': 'keep-alive'})
        self.last_status: Optional[int] = None
        self.last_content: Optional[Dict] = None

    def __str__(self) -> str:
        return f"Connect ({self.protocol}://{self.ip}:{self.port})"

    # --------------------------------------------------------------------------------

    @staticmethod
    def _handle_response(response: requests.Response) -> Optional[Dict]:
        """
        :return: JSON response as dict if there is content, None otherwise
        """
        if len(response.content) == 0:
            return None
        else:
            try:
                return response.json()
            except json.JSONDecodeError:
                return {"message": response.content.decode("utf-8")}

    # noinspection DuplicatedCode
    def _get(self, endpoint: str) -> Optional[Dict]:
        """
        Perform a GET operation for a given endpoint.

        :param endpoint: relative service endpoint, starting with "/"
        :return: response as dict, None if no content
        """
        url = f"{self.protocol}://{self.ip}:{self.port}{endpoint}"
        response = self.session.get(url)
        self.last_status = response.status_code
        self.last_content = self._handle_response(response)
        try:
            response.raise_for_status()
        except Exception as err:
            extra = "" if self.last_content is None else f"\n  -- {self.last_content}"
            raise self.HTTPError(f"GET failed with {err}{extra}")
        return self._handle_response(response)

    # noinspection DuplicatedCode
    def _post(self, endpoint: str, data: Dict) -> Dict:
        """
        Perform a POST operation for a given endpoint.

        :param endpoint: relative service endpoint, starting with "/"
        :param data: body JSON data in dict form
        :return: response as dict, None if no content
        """
        url = f"{self.protocol}://{self.ip}:{self.port}{endpoint}"
        response = self.session.post(url, data=json.dumps(data))
        self.last_status = response.status_code
        self.last_content = self._handle_response(response)
        try:
            response.raise_for_status()
        except Exception as err:
            extra = "" if self.last_content is None else f"\n  -- {self.last_content}"
            raise self.HTTPError(f"POST failed with {err}{extra}")
        return self._handle_response(response)

    # noinspection DuplicatedCode
    def _put(self, endpoint: str, data: Dict) -> Dict:
        """
        Perform a PUT operation for a given endpoint.

        :param endpoint: relative service endpoint, starting with "/"
        :param data: body JSON data in dict form
        :return: response as dict, None if no content
        """
        url = f"{self.protocol}://{self.ip}:{self.port}{endpoint}"
        message = data if not isinstance(data, Dict) else json.dumps(data)
        response = self.session.put(url, data=message)
        self.last_status = response.status_code
        self.last_content = self._handle_response(response)
        try:
            response.raise_for_status()
        except Exception as err:
            extra = "" if self.last_content is None else f"\n  -- {self.last_content}"
            raise self.HTTPError(f"PUT failed with {err}{extra}")
        return self._handle_response(response)

    # noinspection DuplicatedCode
    def _delete(self, endpoint: str) -> Dict:
        """
        Perform a DELETE operation for a given endpoint.

        :param endpoint: relative service endpoint, starting with "/"
        :return: response as dict, None if no content
        """
        url = f"{self.protocol}://{self.ip}:{self.port}{endpoint}"
        response = self.session.delete(url)
        self.last_status = response.status_code
        self.last_content = self._handle_response(response)
        try:
            response.raise_for_status()
        except Exception as err:
            extra = "" if self.last_content is None else f"\n  -- {self.last_content}"
            raise self.HTTPError(f"DELETE failed with {err}{extra}")
        return self._handle_response(response)
