#  coding: utf-8
#  Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved
################################################################################
import json
import logging
import re
import time
from datetime import datetime, timezone, timedelta
from typing import Type, Union, Dict

import _pytest.outcomes
import _pytest.outcomes
import pytest

from lib.connect.hello_connect import HelloConnect

_logger = logging.getLogger(__name__)

__all__ = ["TestCommon"]


################################################################################
# Test class base object
################################################################################

class TestCommon(object):
    """
    Base class for all tests, containing common use methods.
    """

    INVOKED = time.strftime("%Y%m%d-%H%M%S", time.localtime(time.time()))
    PYTEST_FAIL = _pytest.outcomes.Failed

    # noinspection PyUnusedFunction
    @classmethod
    def setup_class(cls):
        _logger.info("\n" + "=" * 80 + "\n" + f"  Executing {cls.__name__} ({cls.__module__})\n" + "=" * 80)

    # noinspection PyUnusedFunction
    @classmethod
    def teardown_class(cls):
        pass

    class Timer(object):
        """
        Context block object that can be used to time operations.
        """

        # noinspection PyUnusedLocal
        def __init__(self, tag: str = None):
            self._start = time.time()
            self._tag = "" if tag is None else tag + " "

        # noinspection PyUnusedLocal
        def __enter__(self):
            self._start = time.time()
            return self

        # noinspection PyUnusedLocal
        def __exit__(self, exc_type: Type[Exception], exc_val: Exception, exc_tb):
            _logger.info(f"{self._tag}TIME: {time.time() - self._start:.2f}s")

    # ----- Test assistance methods -----------------------------------------------------------------

    @staticmethod
    def verify_response(connect: HelloConnect, return_code: int, message: str = None) -> None:
        """
        Quick validate of last response.  If there has been no activity, nothing is checked.

        :param connect: last connection used
        :param return_code: expected http response code
        :param message: expected message, if code is correct (optional)
        """
        if not isinstance(connect, HelloConnect):
            raise TypeError("`connect` must be a HintControlConnect() object!")

        # Check return code
        if connect.last_status is None:  # no operation yet
            return
        if connect.last_status != return_code:
            pytest.fail(f"Saw return_code {connect.last_status} instead of expected {return_code}")

        # Return code ok, check message (if specified)
        if message is not None:
            if connect.last_content is None:
                pytest.fail("We expected response content, but none was provided!")
            if isinstance(connect.last_content, Dict):
                if "message" not in connect.last_content:
                    pytest.fail("Response content was dict, but did not have expected 'message' entry!")
                if message not in connect.last_content['message']:
                    pytest.fail(
                        f"Response content message did not match `{message}`\n  {connect.last_content['message']}")
            else:
                pytest.fail(f"Response content was not in dict format:\n  {connect.last_content}")

    # ----- Other common use methods -----------------------------------------------------------------

    # noinspection PyUnusedFunction
    @staticmethod
    def from_json(data: str) -> dict:
        """Convert JSON information from a source into a more accessible dict object.

        :param data: string containing JSON text
        :return: dict object
        :exception: JSONDecodeError
        :exception: TypeError
        """
        return json.loads(data)

    # noinspection PyUnusedFunction
    @staticmethod
    def to_json(data: dict) -> str:
        """Convert dict into a JSON format for later use.

        :param data: dict to be converted
        :return: string of JSON text
        :exception: TypeError
        """
        return json.dumps(data, sort_keys=True)

    @staticmethod
    def unique(prefix: str = "") -> str:
        """Generate a unique string based on current time.

        :param prefix: (optional) prefix
        :return: Unique time.
        """
        if not isinstance(prefix, str):
            raise TypeError("If specified, `prefix` must be a string!")

        letters = list("ABCDEFGHIJKLMNOPQRSTUVWXYZ")

        now = int(time.time() * 10000)  # to the 1/10th millisecond
        time.sleep(0.0002)  # added to ensure unique time if called repeatedly
        tag = ""
        for _ in range(3):
            tag = letters[int(now % 26)] + tag
            now = now / 26
        return f"{prefix}{tag}{int(now)}"

    @staticmethod
    def uid() -> int:
        """Create unique id based on current time, to the millisecond."""
        time.sleep(0.001)  # assure unique
        return int(time.time() * 1000.0)

    @staticmethod
    def stamp(delta: int = 0) -> datetime:
        """generate a quick timestamp based on current time, to the nearest millisecond

        :param delta: time variance in minutes (default 0)
        """
        use_datetime = datetime.now(timezone.utc) + timedelta(minutes=delta)
        try:
            ms = round(max(0, min(use_datetime.microsecond, 999999)), -3)
            return use_datetime.replace(microsecond=ms)
        except Exception as err:
            _logger.error(f"micro={use_datetime.microsecond} -> {err}")
        return use_datetime

    def is_now(self, check: Union[datetime, str], delta: int = 5) -> bool:
        """
        Verify that supplied time is "now" within +/- some wiggle room in seconds.  Designed to take
        a datetime, or a string in iso format (2023-03-28T13:14:26.460021-04:00) or UTC format
        (2023-03-28T15:52:49.619770769Z).

        NOTE: in the case of UTC format, we need to remove nanoseconds, if seen

        :param check: time to be checked (datetime or postgres string format)
        :param delta: wiggle room in seconds (default 5)
        :return: True if close enough
        """
        now = self.stamp()
        use_check = check
        if isinstance(check, str):
            try:
                use_check = datetime.fromisoformat(check)
            except ValueError:
                # NOTE: This is a hack until we can get the new container built and we can use dateutil.parse
                p = re.compile(r"^(\S*)\.(\d+)Z$")
                m = p.match(check)
                if m and len(m.group(2)) > 6:
                    check = m.group(1) + "." + m.group(2)[0:6] + "Z"
                    _logger.info(f"Trimmed nanoseconds in {check} (.{m.group(2)}Z)")
                use_check = datetime.strptime(check, "%Y-%m-%dT%H:%M:%S.%fZ")
        diff = now.timestamp() - use_check.timestamp()
        return abs(diff) <= delta

    @staticmethod
    def epoch() -> datetime:
        """generate an timestamp for the epoch date (0)."""
        return datetime.fromtimestamp(0, tz=timezone.utc)

    def pinned(self, minutes: int = 0):
        """Generate a fixed timestamp of known value (used primarily for internal testing).

        :param minutes: minutes index (default 0)
        """
        return datetime.replace(self.stamp(), year=1999, month=9, day=13, hour=0, minute=0, second=0,
                                microsecond=0) + timedelta(minutes=minutes)
