#!/usr/bin/env python
# Copyright (c) 2023 VMware Carbon Black, all rights reserved
################################################################################

import argparse
import datetime
import logging
import os
import sys
import time
import uuid
from contextlib import contextmanager
from enum import Enum
from typing import Generator, List, Dict

import requests
from cbd_support.server_api import helpers as server_helpers
from cbd_support.server_api.front_end.front_end_factory import FrontEndFactory
from csp_cbc_auth.interfaces.cbc import CBC
from psc_orgkey import decode_orgkey
from python_sensor_sim import PythonSensorSim
from python_sensor_sim.enums import SensorDeviceType, DeploymentType
from python_sensor_sim.python_sensor_sim import Hints

# noinspection PyUnresolvedReferences
requests.packages.urllib3.disable_warnings()

logger = logging.getLogger(__name__)


class Stats(Enum):
    """
    Enumeration of statistic tracking keys.

    """
    HELLO_CALLS = 0
    LAST_UPDATE = 1
    LAST_LIVEQUERY = 2
    UPDATE_CHANGES = 3
    LIVEQUERY_CHANGES = 4


@contextmanager
def org_user() -> Generator[FrontEndFactory, None, None]:
    """
    Creates an instance of Front End apis through the Node Gateway using CSR access or CBC token access.

    :yields: front_end_api requests session
    """
    use_legacy = True
    if os.getenv("CSP_AUTH_CLIENT_ID", None) is not None or os.getenv("CSP_AUTH_CLIENT_SECRET", None) is not None or \
            os.getenv("CSP_USER_API_TOKEN", None) is not None:
        use_legacy = False
    api_session = None
    if use_legacy:  # older environments
        api_session = server_helpers.create_auth_session()
        logger.info(f"Authorizing with user `{api_session.data['username']}` against {api_session.gateway_url}")
        factory = FrontEndFactory(api_session)  # type: ignore
    else:  # newer mosaic2 environments
        cbc = CBC.auth_oauth_client_credentials()
        factory = FrontEndFactory(cbc.cbc_session)
    yield factory
    if api_session is not None:
        api_session.close()


@contextmanager
def sensor_sim(user_session: FrontEndFactory, device_type: SensorDeviceType,
               deployment: DeploymentType, prefix: str = None) -> Generator[PythonSensorSim, None, None]:
    """
    Creates an instance of python sensor sim for the provided user, device type and deployment type.

    FUTURE: add ability to add sensor in different orgs besides 1?

    :param user_session: FrontEndFactory
    :param device_type: SensorDeviceType
    :param deployment: DeploymentType
    :param prefix: Special identification prefix
    :return Generator[PythonSensorSim, None, None]
    """
    use_prefix = "" if prefix is None else f"{prefix}-"
    device = PythonSensorSim(org_user=user_session, device_type=device_type, deployment_type=deployment,
                             device_name=f'{use_prefix}enki-hello-{uuid.uuid4()}',
                             email=f"{use_prefix}enki-hello@vmware.com")
    try:
        yield device
    finally:
        logger.info("... Done\n\n" + "=" * 80)
        device.uninstall_and_delete_device()


def device_state_banner(device, tag: str = None, hints: List[Hints] = None, stats: Dict = None) -> None:
    """
    Show the current sensor state.
    :param device: device object
    :param tag: banner identification
    :param hints: hints returned from latest hello call
    :param stats: dictionary with statistics information
    """
    use_tag = "" if tag is None else f"===== {tag} ==========\n"
    use_hints = ""
    if hints is not None:
        hint_names = [x.name for x in hints]
        hints_str = "\n                  ".join(hint_names)
        use_hints = f"\n           Hints: {hints_str}"

    # NOTE: Times returned are milliseconds after epoch time which must be converted
    lut = "" if device.last_update_time is None else f"  ({datetime.datetime.utcfromtimestamp(device.last_update_time / 1000)})"
    llqt = "" if device.last_live_query_time is None else f"  ({datetime.datetime.utcfromtimestamp(device.last_live_query_time / 1000)})"

    lut_changed = ""
    if hasattr(device, "last_update_time_saved") and device.last_update_time > device.last_update_time_saved:
        lut_changed = "   <<< Updated"
        if stats is not None:
            stats[Stats.UPDATE_CHANGES] = stats.get(Stats.UPDATE_CHANGES, 0) + 1
            stats[Stats.LAST_UPDATE] = device.last_update_time
    llqt_changed = ""
    if hasattr(device,
               "last_live_query_time_saved") and device.last_live_query_time is not None and device.last_live_query_time_saved is not None:
        if device.last_live_query_time > device.last_live_query_time_saved:
            llqt_changed = "   <<< Updated"
            if stats is not None:
                stats[Stats.LIVEQUERY_CHANGES] = stats.get(Stats.LIVEQUERY_CHANGES, 0) + 1
                stats[Stats.LAST_LIVEQUERY] = device.last_live_query_time
    elif (not hasattr(device,
                      "last_live_query_time_saved") or device.last_live_query_time_saved is None) and device.last_live_query_time is not None:
        llqt_changed = "   <<< Updated"
        if stats is not None:
            stats[Stats.LIVEQUERY_CHANGES] += 1

    # NOTE: This utility uses private attributes "last_update_time_saved" and "last_live_query_time_saved" for
    #       time change flagging; these are created/updated with the following method
    timestamp_change_tracking_update(device)

    eo = decode_orgkey(device.org_key)
    logging.info(
        f"{use_tag}Sensor: Hostname {device.device_host_name}, Orgkey {device.org_key} (env {eo.env}, id {eo.org_id})\n"
        f"           DeploymentType: {device.deployment_type}\n"
        f"           DeviceType:     {device.device_type}\n"
        f"           Last Update:    {device.last_update_time}{lut}{lut_changed}\n"
        f"           Live Query:     {device.last_live_query_time}{llqt}{llqt_changed}{use_hints}\n")


def timestamp_change_tracking_update(dev_obj) -> None:
    """
    Stash the current timestamp info for the purpose of flagging updates.

    :param dev_obj: device object
    """
    dev_obj.last_update_time_saved = dev_obj.last_update_time
    if dev_obj.last_live_query_time is not None or not hasattr(dev_obj, "last_live_query_time_saved"):
        dev_obj.last_live_query_time_saved = dev_obj.last_live_query_time


################################################################################
# Setup functions
################################################################################


class BlankLinesHelpFormatter(argparse.HelpFormatter):
    """
    Add empty line if help ends with \n
    """

    def _split_lines(self, text: str, width: int):
        # noinspection PyProtectedMember
        lines = super()._split_lines(text, width)
        if text.endswith('\n'):
            lines += ['']
        return lines


def build_cli_parser() -> argparse.Namespace:
    """
    Build the set of accepted options.
    """
    # noinspection PyTypeChecker
    parser = argparse.ArgumentParser(description="Perform a device checkin to an environment's hello service.",
                                     formatter_class=BlankLinesHelpFormatter)

    # for each supported output type, add an option
    parser.add_argument("--debug", action="store_true", dest="debug", default=False,
                        help="enable debug logging\n")

    parser.add_argument("--age", action="store", dest="age", type=int, default=None,
                        help="age of the sensor since the last checkin (in minutes)")
    parser.add_argument("--count", action="store", dest="count", type=int, default=1,
                        help="number of requests to make (default is 1)")
    parser.add_argument("--delay", action="store", dest="delay", type=int, default=10,
                        help="delay in seconds between requests when count is greater than 1 (default is 10)")
    parser.add_argument("--deploy", action="store", dest="deployment", default="endpoint",
                        help="deployment type, one of 'endpoint' or 'workload' (default is 'endpoint')")
    parser.add_argument("--info", action="store_true", dest="info", default=False,
                        help="dump information about the user and simulated sensor as well as the config")
    parser.add_argument("--prefix", action="store", dest="prefix", default=None,
                        help="identifying prefix added to hostname and user email")
    parser.add_argument("--stats", action="store_true", dest="stats", default=False,
                        help="output operational statistics after the run")
    parser.add_argument("--type", action="store", dest="devtype", default="windows",
                        help="device type, one of 'windows', 'mac', 'linux', 'android' or 'ios' (default is 'windows')")

    # FUTURE: Add ability to control client version in PythonSensorSim, if this proves useful

    args = parser.parse_args()
    args.devtype = args.devtype.lower().strip()
    if args.devtype not in ["windows", "mac", "linux", "android", "ios"]:
        raise RuntimeError("--type must be one of 'windows', 'mac', 'linux', 'android' or 'ios'")
    args.deployment = args.deployment.lower().strip()
    if args.deployment not in ["endpoint", "workload"]:
        raise RuntimeError("--deploy must be one of 'endpoint' or 'workload'")
    args.count = max(args.count, 1)
    args.delay = max(args.delay, 0)

    return args


def dump_config():
    """
    Show current configuration being used (minus private information)
    """
    logger.info(f"CBC_CSR_URL: {os.getenv('CBC_CSR_URL')}")
    logger.info(f"CSP_ENV: {os.getenv('CSP_ENV')}")
    logger.info(f"CBC_ORG_URL: {os.getenv('CBC_ORG_URL')}")
    logger.info(f"CSP_AUTH_CLIENT_ID: {os.getenv('CSP_AUTH_CLIENT_ID')}")
    logger.info(f"CSP_AUTH_CLIENT_SECRET: {os.getenv('CSP_AUTH_CLIENT_SECRET')}")
    logger.info(f"CSP_USER_API_TOKEN: {os.getenv('CSP_USER_API_TOKEN')}")
    logger.info(f"CSR_ADMIN: {os.getenv('CSR_ADMIN')}")
    logger.info(f"CSR_PASSWORD: {os.getenv('CSR_PASSWORD')}")
    logger.info(f"CSR: {os.getenv('CSR')}")
    logger.info(f"DASHBOARD: {os.getenv('DASHBOARD')}")
    logger.info(f"DEVICE_GATEWAY_URL: {os.getenv('DEVICE_GATEWAY_URL')}")
    logger.info(f"DEVICES: {os.getenv('DEVICES')}")
    logger.info(f"INTSVCS: {os.getenv('INTSVCS')}")
    logger.info(f"ORG_ADMIN_PASSWORD: {os.getenv('ORG_ADMIN_PASSWORD')}")
    logger.info(f"ORG_ADMIN: {os.getenv('ORG_ADMIN')}")
    logger.info(f"TEST_ENV: {os.getenv('TEST_ENV')}")


################################################################################
# Begin Main
################################################################################


def main():
    try:
        args = build_cli_parser()
    except Exception as err:
        logger.error(f"Failed setup: {err}")
        sys.exit(1)
    level = logging.DEBUG if args.debug else logging.INFO
    logging.getLogger().setLevel(level)
    fmt = '%(asctime)s - [%(process)d] %(levelname)7s [%(name)s:%(lineno)d] %(message)s'
    logging.basicConfig(level=level, format=fmt)
    if args.info:
        dump_config()

    devtype = SensorDeviceType.WINDOWS
    if args.devtype == "mac":
        devtype = SensorDeviceType.MAC
    elif args.devtype == "linux":
        devtype = SensorDeviceType.LINUX
    deploy = DeploymentType.ENDPOINT
    if args.devtype == "workload":
        deploy = DeploymentType.WORKLOAD

    # enabling statistics
    if args.stats:
        stats_data = {Stats.LAST_LIVEQUERY: 0, Stats.LIVEQUERY_CHANGES: 0}  # Initialize these since there might be none
    else:
        stats_data = None

    try:
        with org_user() as user_session:
            if args.info:
                if hasattr(user_session.api_session, "org"):
                    org_name = user_session.api_session.org.name
                else:
                    org_id = user_session.api_session.user_info.get('currentOrgId')
                    org_name = user_session.api_session.user_info.get('orgNames').get(str(org_id))
                data = "\n" + "=" * 80 + "\n" + \
                       f"  USER: {user_session.api_session.login_id} on {user_session.api_session.gateway_url}\n" + \
                       f"     EMAIL: {user_session.api_session.email}\n" + \
                       f"  ORG: {user_session.api_session.org_key} (id {user_session.api_session.org_id}, name '{org_name}')\n" + \
                       "=" * 80 + "\n"
                logger.info(data)

                with sensor_sim(user_session=user_session, device_type=devtype, deployment=deploy,
                                prefix=args.prefix) as device:
                    # NOTE: Uncomment and debug if you are interested in the device policies
                    # policy = device.request_policy()

                    if args.info:
                        dev_info = device.get_app_services_device_info(False)
                        data = "\n" + "=" * 80 + "\n" + \
                               f"  HOSTNAME: {dev_info['deviceName']} (Dev ID {dev_info['deviceId']})\n" + \
                               f"     TYPE:       {dev_info['deviceType']}\n" + \
                               f"     DEPLOYMENT: {device.deployment_type.name}\n" + \
                               f"  GROUP:    {dev_info['group']} (Grp ID {dev_info['groupId']})\n" + \
                               f"  ORGKEY:   {device.org_key}\n" + \
                               f"  COMPANY:  {device.company_code}\n" + \
                               f"  STATUS:   {dev_info['status']}, REGISTERED TIME: {datetime.datetime.fromtimestamp(dev_info['registeredTime'] / 1000)}\n" + \
                               "=" * 80 + "\n"
                        logger.info(data)
                    device_state_banner(device, "Created", None)
                    use_delay = False
                    if args.age is not None:
                        now = int(time.time() * 1000)
                        device.last_update_time = now - (args.age * 60000)
                        device.last_live_query_time = now - (args.age * 60000)

                    for i in range(args.count):
                        if use_delay and args.count > 0 and args.delay > 0:
                            time.sleep(args.delay)
                        use_delay = True

                        # NOTE: Bobby Speaker, re: sensor workflow
                        #   * Sensor sends hello and includes its LastLQTimestamp in request
                        #   * Response to hello includes backend provided time but sensor ignores it. Hello response also
                        #     includes hints of what actions for sensor to take
                        #   * One of those hints tells sensor that there are more LQ's to run.    If that hint is present,
                        #     sensor will send a separate message to /services/psc/v1/livequery/ to get the list of LQ's to run
                        #   * Once those are enqueued, we update our timestamp that will go into the next hello request so
                        #     that backend knows not to send that hint again unless it has queries queued after the last
                        #     time we downloaded LQ requests

                        # NOTE: Since PythonSensorSim cannot (presently) make LQ calls, we simulate this by forcing the
                        #       use of the last_live_query_time value last returned
                        hints = device.send_hello_request(
                            override_last_live_query_time=device.last_live_query_time_saved)
                        if stats_data is not None:
                            stats_data[Stats.HELLO_CALLS] = stats_data.get(Stats.HELLO_CALLS, 0) + 1
                        device_state_banner(device, "After Hello", hints, stats_data)

                    time.sleep(5)  # to allow for cleanup
    except KeyboardInterrupt:
        logger.warning(" ... aborted by user!")
    except Exception as err:
        logger.error(f" ... aborted due to problem: {err}")

    if args.stats and stats_data is not None:
        print("\n===== Execution Statistics ========================================")
        print(f"  {stats_data[Stats.HELLO_CALLS]} hello calls")
        print(
            f"    Last Update Time:    {stats_data[Stats.LAST_UPDATE]:13d} ({stats_data[Stats.UPDATE_CHANGES]} changes)")
        print(
            f"    Last Livequery Time: {stats_data[Stats.LAST_LIVEQUERY]:13d} ({stats_data[Stats.LIVEQUERY_CHANGES]} changes)")


if __name__ == '__main__':
    main()
